"""Crypto wrapper implementations.

A crypto wrapper encapsulates all cryptographic operations that have to be
applied to node representations before they can be safely inserted into the
(untrusted) database, e.g., encryption and authentication. It defines the keys
(digests) under which (wrapped) node representations are stored and specifies
how node representations can be extracted from (digest, value) pairs stored in
the database.

This module includes several crypto wrapper implementations with different
security properties.
"""
import abc
import hashlib
import hmac
import struct
import logging


# format strings to encode parameters
FORMAT_HEIGHT = 'i'
FORMAT_ISROOT = '?'
FORMAT_HEIGHT_ISROOT = FORMAT_HEIGHT + FORMAT_ISROOT


class IntegrityError(ValueError):
    """Raised if an integrity verification fails."""
    pass


class AuthenticityError(IntegrityError):
    """Raised if an authenticity verification fails."""
    pass


class BaseCryptoWrapper(object):

    """Abstract class specifying the crypto wrapper interface."""
    __metaclass__ = abc.ABCMeta
    __slots__ = []

    """Defines the constant size of digests generated by this crypto wrapper."""
    DIGEST_SIZE = None

    def wrap_value(self, value, height, is_root):
        """Converts a node representation into a (digest, value) pair that can
        be safely inserted into the database.

        Args:
            value (str): The node representation.
            height (int): The height of the node in its chunk tree.
            is_root (bool): Whether or not the node is the chunk tree's root.

        Returns:
            tuple: (wrapped_value, digest).
        """
        raise NotImplementedError

    def unwrap_value(self, value, digest, height, is_root, length=-1):
        """Converts the representation of a node as stored in the database
        (e.g., an encrypted representation) into its `normal`, e.g., decrypted,
        representation.

        Args:
            value (str): Wrapped node representation.
            digest (str): Key under which the node is stored in the database.
            height (int): Height of the node in its chunk tree.
            is_root (bool): Whether or not the node is the chunk tree's root.
            length [Optional(int)]: Length of the content represented by this
                node. Defaults to -1.

        Returns:
            str: Node representation.
        """
        raise NotImplementedError


HASHLIB_SHA256 = hashlib.sha256


class SHA_256(BaseCryptoWrapper):

    """SHA-256 crypto wrapper.

    Provides integrity for chunk tree nodes.

    Nodes are represented as follows:
        * value: <value>
        * digest: SHA-256(<value>)
    """

    __slots__ = []

    DIGEST_SIZE = 32

    def __init__(self):
        super(SHA_256, self).__init__()

    def wrap_value(self, value, height, is_root):
        """Uses SHA-256 hash of node representation as digest and value as is.

        See :meth:`.BaseCryptoWrapper.wrap_value`.
        """
        m = HASHLIB_SHA256()
        m.update(struct.pack(FORMAT_HEIGHT, height))
        m.update(value)
        return value, m.digest()

    def unwrap_value(self, value, digest, height, is_root, length=-1):
        """Verifies SHA-256 hash and returns node representation on success.

        Raises:
            IntegrityError: If digest does not match.

        See :meth:`.BaseCryptoWrapper.unwrap_value`.
        """
        m = HASHLIB_SHA256()
        m.update(struct.pack(FORMAT_HEIGHT, height))
        m.update(value)
        if m.digest() != digest or (height == 0 and len(value) < length):
            raise IntegrityError
        return value


HMAC_new = hmac.new


class HMAC_SHA_256(BaseCryptoWrapper):

    """HMAC-SHA-256 crypto wrapper.

    Provides authenticity for chunk tree nodes based on a symmetric 32-bytes key
    specified during instantiation.

    Root nodes are handled identically to inner nodes at the same level.

    Nodes are represented as follows:
        * value: <value>
        * digest: HMAC-SHA-256(<key>, <height> || <value>)

    Args:
        key (str): Cryptographic key used for symmetric authentication.
    """

    __slots__ = ['_key']

    DIGEST_SIZE = 32

    def __init__(self, key):
        super(HMAC_SHA_256, self).__init__()
        self._key = key

    def wrap_value(self, value, height, is_root):
        """Uses SHA-256-based HMAC of node representation and height as digest
        and value as is.

        See :meth:`.BaseCryptoWrapper.wrap_value`.
        """
        m = HMAC_new(self._key, digestmod=HASHLIB_SHA256)
        m.update(struct.pack(FORMAT_HEIGHT, height))
        m.update(value)
        return value, m.digest()

    def unwrap_value(self, value, digest, height, is_root, length=-1):
        """Verifies SHA-256-based HMAC and returns node representation on
        success.

        Raises:
            AuthenticityError: If digest does not match.

        See :meth:`.BaseCryptoWrapper.unwrap_value`.
        """
        m = HMAC_new(self._key, digestmod=HASHLIB_SHA256)
        m.update(struct.pack(FORMAT_HEIGHT, height))
        m.update(value)
        if m.digest() != digest or (height == 0 and len(value) < length):
            raise AuthenticityError
        return value


class HMAC_SHA_256_DISTINGUISHED_ROOT(HMAC_SHA_256):

    """HMAC-SHA-256 crypto wrapper with distinguished root representation.

    Provides authenticity for chunk tree nodes based on a symmetric 32-bytes key
    specified during instantiation.

    Root nodes are handled differently from inner nodes at the same level.

    Nodes are represented as follows:
        * value: <value>
        * digest: HMAC-SHA-256(<key>, <height> || <is_root> || <value>)

    Args:
        key (str): Cryptographic key used for symmetric authentication.
    """

    __slots__ = []

    def __init__(self, key):
        HMAC_SHA_256.__init__(self, key)

    def wrap_value(self, value, height, is_root):
        """Uses SHA-256-based HMAC of node representation, height and is_root
        flag as digest and value as is.

        See :meth:`.BaseCryptoWrapper.wrap_value`.
        """
        m = HMAC_new(self._key, digestmod=HASHLIB_SHA256)
        m.update(struct.pack(FORMAT_HEIGHT_ISROOT, height, is_root))
        m.update(value)
        return value, m.digest()

    def unwrap_value(self, value, digest, height, is_root, length=-1):
        """Verifies SHA-256-based HMAC and returns node representation on
        success.

        Raises:
            AuthenticityError: If digest does not match.

        See :meth:`.BaseCryptoWrapper.unwrap_value`.
        """
        m = HMAC_new(self._key, digestmod=HASHLIB_SHA256)
        m.update(struct.pack(FORMAT_HEIGHT_ISROOT, height, is_root))
        m.update(value)
        if m.digest() != digest or (height == 0 and len(value) < length):
            raise AuthenticityError
        return value


# HMAC_SHA_256_R = HMAC_SHA_256.DIGEST_SIZE + 8
# 
# 
# class HMAC_SHA_256_DISTINGUISHED_ROOT_WITH_LEAF_PADDING(HMAC_SHA_256):
#  
#     """HMAC-SHA-256 crypto wrapper with distinguished root representation and
#     leaf padding.
#  
#     Provides authenticity for chunk tree nodes based on a symmetric 32-bytes key
#     specified during instantiation.
#  
#     Root nodes are handled differently from inner nodes at the same level.
#     
#     WARNING: not yet supported by SecCS!
#  
#     Nodes are represented as follows:
#         * value: <value>
#         * digest: HMAC-SHA-256(<key>, <height> || <is_root> || <value>)
#  
#     Args:
#         key (str): Cryptographic key used for symmetric authentication.
#     """
#      
#     __slots__ = []
#  
#     def __init__(self, key):
#         HMAC_SHA_256.__init__(self, key)
#  
#     def wrap_value(self, value, height, is_root):
#         """Pads leaf node representations with zero bytes to make them
#         indifferentiable from superchunks and uses SHA-256-based HMAC of the
#         resulting node representation, height and is_root flag as digest.
#  
#         See :meth:`.BaseCryptoWrapper.wrap_value`.
#         """
#         m = HMAC_new(self._key, digestmod=HASHLIB_SHA256)
#         m.update(struct.pack(FORMAT_HEIGHT_ISROOT, height, is_root))
#         if height == 0:
#             value += b'\x00' * (HMAC_SHA_256_R - (len(value) % HMAC_SHA_256_R))
#         m.update(value)
#         return value, m.digest()
#  
#     def unwrap_value(self, value, digest, height, is_root, length=-1):
#         """Verifies SHA-256-based HMAC, removes padding and returns resulting
#         node representation on success.
#  
#         Raises:
#             AuthenticityError: If digest does not match.
#  
#         See :meth:`.BaseCryptoWrapper.unwrap_value`.
#         """
#         m = HMAC_new(self._key, digestmod=HASHLIB_SHA256)
#         m.update(struct.pack(FORMAT_HEIGHT_ISROOT, height, is_root))
#         m.update(value)
#         if m.digest() != digest or (height == 0 and len(value) < length):
#             raise AuthenticityError
#         return value[:length] if height == 0 and length > -1 else value


try:
    from Crypto.Cipher import AES
    AES_MODE_SIV = AES.MODE_SIV
except ImportError:
    logging.getLogger(__name__).warn('PyCrypto is not available, disabling AES wrappers.')
except AttributeError:
    logging.getLogger(__name__).warn('Your PyCrypto version is too old, disabling '
                                     'AES-SIV wrappers which depend on PyCrypto >= 2.7a1.')
else:
    AES_new_fn = AES.new
    
    
    class AES_SIV_256(BaseCryptoWrapper):
    
        """AES-SIV-256 crypto wrapper.
    
        Provides confidentiality and authenticity for chunk tree nodes based on a
        symmetric 32-bytes key specified during instantiation.
    
        Root nodes are handled identically to inner nodes at the same level.

        Nodes are represented as follows:
            * value: AES-SIV-256(<key>, <value>, additional_data=<height>)
            * digest: <digest produced by AES-SIV-256>
    
        Args:
            key (str): Cryptographic key used for symmetric encryption and
                authentication.

        Note:
            Requires PyCrypto >= 2.7a1.
        """
    
        __slots__ = ['_key', '_zero_digest']
    
        DIGEST_SIZE = 16
    
        def __init__(self, key):
            super(AES_SIV_256, self).__init__()
            self._key = key
    
            """The AES-SIV implementation does not support encryption of empty
            strings, so we represent empty-string nodes by a designated zero_digest
            instead.
            """
            cipher = AES_new_fn(self._key, AES_MODE_SIV)
            cipher.update(b'empty_additional_data')
            _, self._zero_digest = cipher.encrypt_and_digest(b'empty_string')
            self._zero_digest = b'\x00' * AES_SIV_256.DIGEST_SIZE
    
        def wrap_value(self, value, height, is_root):
            """Encrypts node representation using deterministic authenticated
            encryption, i.e., with AES in SIV mode, including node height as
            additional data that is authenticated, resulting in a digest (MAC)
            that is used as `digest` and a ciphertext that is used as `value`.
    
            Note:
                As AES-SIV cannot encrypt empty contents, a distinguished zero
                digest is artifically assigned to empty node representations
                instead.
    
            See :meth:`.BaseCryptoWrapper.wrap_value`.
            """
            if value == b'':
                return value, self._zero_digest
            cipher = AES_new_fn(self._key, AES_MODE_SIV)
            cipher.update(struct.pack(FORMAT_HEIGHT, height))
            return cipher.encrypt_and_digest(value)
    
        def unwrap_value(self, value, digest, height, is_root, length=-1):
            """Decrypts and verifies node representation and returns the result on
            success.
    
            Raises:
                AuthenticityError: If digest does not match.
    
            See :meth:`.BaseCryptoWrapper.unwrap_value`.
            """
            if value == b'' and digest == self._zero_digest:
                return value
            cipher = AES_new_fn(self._key, AES_MODE_SIV)
            cipher.update(struct.pack(FORMAT_HEIGHT, height))
            try:
                return cipher.decrypt_and_verify(value, digest)
            except ValueError as e:
                raise AuthenticityError(e)
    
    
    class AES_SIV_256_DISTINGUISHED_ROOT(AES_SIV_256):
    
        """AES-SIV-256 crypto wrapper with distinguished root representation.
    
        Provides confidentiality and authenticity for chunk tree nodes based on a
        symmetric 32-bytes key specified during instantiation.
    
        Root nodes are handled differently from inner nodes at the same level.

        Nodes are represented as follows:
            * value: AES-SIV-256(<key>, <value>, additional_data=<height>||<is_root>)
            * digest: <digest produced by AES-SIV-256>
    
        Args:
            key (str): Cryptographic key used for symmetric encryption and
                authentication.

        Note:
            Requires PyCrypto >= 2.7a1.
        """

        __slots__ = []
            
        def __init__(self, key):
            AES_SIV_256.__init__(self, key)
    
        def wrap_value(self, value, height, is_root):
            """Encrypts node representation using deterministic authenticated
            encryption, i.e., with AES in SIV mode, including node height and
            is_root flag as additional data that is authenticated, resulting in a
            digest (MAC) that is used as `digest` and a ciphertext that is used as
            `value`.
    
            See :meth:`.AES_SIV_256.wrap_value`.
            """
            if value == b'':
                return value, self._zero_digest
            cipher = AES_new_fn(self._key, AES_MODE_SIV)
            cipher.update(struct.pack(FORMAT_HEIGHT_ISROOT, height, is_root))
            return cipher.encrypt_and_digest(value)
    
        def unwrap_value(self, value, digest, height, is_root, length=-1):
            """Decrypts and verifies node representation and returns the result on
            success.
    
            Raises:
                AuthenticityError: If digest does not match.
    
            See :meth:`.BaseCryptoWrapper.unwrap_value`.
            """
            if value == b'' and digest == self._zero_digest:
                return value
            cipher = AES_new_fn(self._key, AES_MODE_SIV)
            cipher.update(struct.pack(FORMAT_HEIGHT_ISROOT, height, is_root))
            try:
                return cipher.decrypt_and_verify(value, digest)
            except ValueError as e:
                raise AuthenticityError(e)



"""Unit tests.

Performs tests of the seccs data structure with randomly generated contents.

If run via tox / py.test, a single (random) run is performed.

If run directly, the tests are repeated with all seeds that triggered errors in
the past. Note that this requires an extensive amount of time!
"""

import functools
import math
import os
import sys
import unittest

import seccs


# ensure that we include from the parent path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))


class RCTest(unittest.TestCase):

    def test_base_rc(self):
        class Test(seccs.rc.BaseReferenceCounter):

            def inc(self, *args, **kwargs):
                return super(Test, self).inc(*args, **kwargs)

            def dec(self, *args, **kwargs):
                return super(Test, self).dec(*args, **kwargs)

        rc = Test()
        key = os.urandom(32)
        self.assertRaises(NotImplementedError, rc.inc, key)
        self.assertRaises(NotImplementedError, rc.dec, key)
        self.assertRaises(NotImplementedError, rc.get, key)

    def test_no_rc(self):
        rc = seccs.rc.NoReferenceCounter()
        key = os.urandom(32)
        self.assertEqual(rc.inc(key), 1)
        self.assertEqual(rc.dec(key), 1)
        self.assertEqual(rc.get(key), 1)

    def test_database_rc(self):
        database = dict()

        rc = seccs.rc.DatabaseReferenceCounter(database)

        key = os.urandom(32)

        self.assertNotIn(key, database)
        self.assertEqual(rc.inc(key), 1)
        self.assertIn(key, database)
        self.assertEqual(rc.inc(key), 2)
        self.assertEqual(rc.get(key), 2)
        self.assertEqual(rc.dec(key), 1)
        self.assertEqual(rc.dec(key), 0)
        self.assertEqual(rc.get(key), 0)
        self.assertNotIn(key, database)

    def test_key_suffix_database_rc(self):
        database = dict()
        suffix = os.urandom(2)

        rc = seccs.rc.KeySuffixDatabaseReferenceCounter(database, suffix)

        key = os.urandom(32)

        self.assertNotIn(key + suffix, database)
        self.assertEqual(rc.inc(key), 1)
        self.assertNotIn(key, database)
        self.assertIn(key + suffix, database)
        self.assertEqual(rc.inc(key), 2)
        self.assertEqual(rc.get(key), 2)
        self.assertEqual(rc.dec(key), 1)
        self.assertEqual(rc.dec(key), 0)
        self.assertEqual(rc.get(key), 0)
        self.assertNotIn(key + suffix, database)


class CryptoWrapperTest(unittest.TestCase):

    def test_base_cw(self):
        class Test(seccs.crypto_wrapper.BaseCryptoWrapper):

            def wrap_value(self, *args, **kwargs):
                return super(Test, self).wrap_value(*args, **kwargs)

            def unwrap_value(self, *args, **kwargs):
                return super(Test, self).unwrap_value(*args, **kwargs)

        cw = Test()
        self.assertRaises(NotImplementedError, cw.wrap_value, b'', 0, False)
        self.assertRaises(
            NotImplementedError, cw.unwrap_value, b'', b'', 0, False)

    def test_SHA_256(self):
        cw = seccs.crypto_wrapper.SHA_256()

        value = os.urandom(100)
        height = 0
        is_root = False

        wrapped_value, digest = cw.wrap_value(value, height, is_root)

        self.assertEqual(value, wrapped_value)

        self.assertEqual(
            value, cw.unwrap_value(wrapped_value, digest, height, is_root))

        self.assertRaises(seccs.crypto_wrapper.IntegrityError, cw.unwrap_value,
                          wrapped_value + b'x', digest, height, is_root)
        self.assertRaises(seccs.crypto_wrapper.IntegrityError, cw.unwrap_value,
                          wrapped_value, digest[:-3] + b'xyz', height, is_root)
        self.assertRaises(seccs.crypto_wrapper.IntegrityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, is_root)
        
        cw.unwrap_value(wrapped_value, digest, height, not is_root)
        
        self.assertRaises(seccs.crypto_wrapper.IntegrityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, not is_root)

    def test_HMAC_SHA_256(self):
        key = os.urandom(32)
        cw = seccs.crypto_wrapper.HMAC_SHA_256(key)
        key2 = os.urandom(32)
        cw2 = seccs.crypto_wrapper.HMAC_SHA_256(key2)

        value = os.urandom(100)
        height = 0
        is_root = False

        wrapped_value, digest = cw.wrap_value(value, height, is_root)

        self.assertNotEqual(
            (wrapped_value, digest), cw2.wrap_value(value, height, is_root))

        self.assertEqual(value, wrapped_value)

        self.assertEqual(
            value, cw.unwrap_value(wrapped_value, digest, height, is_root))

        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value + b'x', digest, height, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest[:-3] + b'xyz', height, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, is_root)

        cw.unwrap_value(wrapped_value, digest, height, not is_root)

        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, not is_root)

    def test_HMAC_SHA_256_DISTINGUISHED_ROOT(self):
        key = os.urandom(32)
        cw = seccs.crypto_wrapper.HMAC_SHA_256_DISTINGUISHED_ROOT(key)
        key2 = os.urandom(32)
        cw2 = seccs.crypto_wrapper.HMAC_SHA_256_DISTINGUISHED_ROOT(key2)

        value = os.urandom(100)
        height = 0
        is_root = False

        wrapped_value, digest = cw.wrap_value(value, height, is_root)

        self.assertNotEqual(
            (wrapped_value, digest), cw2.wrap_value(value, height, is_root))

        self.assertEqual(value, wrapped_value)

        self.assertEqual(
            value, cw.unwrap_value(wrapped_value, digest, height, is_root))

        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value + b'x', digest, height, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest[:-3] + b'xyz', height, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height, not is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, not is_root)

#     def test_HMAC_SHA_256_DISTINGUISHED_ROOT_WITH_LEAF_PADDING(self):
#         key = os.urandom(32)
#         cw = seccs.crypto_wrapper.HMAC_SHA_256_DISTINGUISHED_ROOT_WITH_LEAF_PADDING(
#             key)
#         key2 = os.urandom(32)
#         cw2 = seccs.crypto_wrapper.HMAC_SHA_256_DISTINGUISHED_ROOT_WITH_LEAF_PADDING(
#             key2)
# 
#         value = os.urandom(100)
#         height = 0
#         is_root = False
# 
#         wrapped_value, digest = cw.wrap_value(value, height, is_root)
# 
#         self.assertNotEqual(
#             (wrapped_value, digest), cw2.wrap_value(value, height, is_root))
# 
#         self.assertEqual(value, wrapped_value[:len(value)])
#         self.assertNotEqual(len(value), len(wrapped_value))
# 
#         self.assertEqual(
#             value, cw.unwrap_value(wrapped_value, digest, height, is_root, len(value)))
# 
#         self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
#                           wrapped_value + b'x', digest, height, is_root)
#         self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
#                           wrapped_value, digest[:-3] + b'xyz', height, is_root)
#         self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
#                           wrapped_value, digest, height + 1, is_root)
#         self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
#                           wrapped_value, digest, height, not is_root)
#         self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
#                           wrapped_value, digest, height + 1, not is_root)

    def test_AES_SIV_256(self):
        try:
            seccs.crypto_wrapper.AES_SIV_256
        except AttributeError:
            self.skipTest('AES-SIV support is missing.')

        key = os.urandom(32)
        cw = seccs.crypto_wrapper.AES_SIV_256(key)
        key2 = os.urandom(32)
        cw2 = seccs.crypto_wrapper.AES_SIV_256(key2)

        value = os.urandom(100)
        height = 0
        is_root = False

        wrapped_value, digest = cw.wrap_value(value, height, is_root)

        self.assertNotEqual(
            (wrapped_value, digest), cw2.wrap_value(value, height, is_root))

        self.assertNotEqual(value, wrapped_value)
        self.assertEqual(len(value), len(wrapped_value))

        self.assertEqual(
            value, cw.unwrap_value(wrapped_value, digest, height, is_root))

        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value + b'x', digest, height, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest[:-3] + b'xyz', height, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, is_root)

        cw.unwrap_value(wrapped_value, digest, height, not is_root)

        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, not is_root)

        wrapped_value, digest = cw.wrap_value(b'', height, is_root)
        self.assertEqual(
            b'', cw.unwrap_value(wrapped_value, digest, height, is_root))

    def test_AES_SIV_256_DISTINGUISHED_ROOT(self):
        try:
            seccs.crypto_wrapper.AES_SIV_256
        except AttributeError:
            self.skipTest('AES-SIV support is missing.')

        key = os.urandom(32)
        cw = seccs.crypto_wrapper.AES_SIV_256_DISTINGUISHED_ROOT(key)
        key2 = os.urandom(32)
        cw2 = seccs.crypto_wrapper.AES_SIV_256_DISTINGUISHED_ROOT(key2)

        value = os.urandom(100)
        height = 0
        is_root = False

        wrapped_value, digest = cw.wrap_value(value, height, is_root)

        self.assertNotEqual(
            (wrapped_value, digest), cw2.wrap_value(value, height, is_root))

        self.assertNotEqual(value, wrapped_value)
        self.assertEqual(len(value), len(wrapped_value))

        self.assertEqual(
            value, cw.unwrap_value(wrapped_value, digest, height, is_root))

        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value + b'x', digest, height, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest[:-3] + b'xyz', height, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height, not is_root)
        self.assertRaises(seccs.crypto_wrapper.AuthenticityError, cw.unwrap_value,
                          wrapped_value, digest, height + 1, not is_root)

        wrapped_value, digest = cw.wrap_value(b'', height, is_root)
        self.assertEqual(
            b'', cw.unwrap_value(wrapped_value, digest, height, is_root, 0))


class SecCSLiteTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        import random
        self.seed = kwargs.pop('seed', random.random())
        self.random = random.Random(self.seed)

        self.S = kwargs.pop('S', 128)
        self.crypto_wrapper = kwargs.pop('crypto_wrapper', seccs.crypto_wrapper.SHA_256())

        unittest.TestCase.__init__(self, *args, **kwargs)

    def shortDescription(self):
        return 'SecCSLite: S={S},seed=float.fromhex(\'{seed}\')'.format(S=self.S, seed=self.seed.hex())

    def setUp(self):
        self.setUpInstance(seccs.SecCSLite)

    def setUpInstance(self, seccs_class):
        kvs = dict()
        ref_kvs = dict()

        crypto_wrapper = self.crypto_wrapper
        self.digest_size = crypto_wrapper.DIGEST_SIZE

        ''' functions that provide storage consumption information '''
        self.kvs_size = lambda: functools.reduce(
            lambda size, kv: size + len(kv[0]) + len(kv[1]), kvs.items(), 0)
        self.kvs_count = lambda: len(kvs)
        self.rc_count = lambda: len(ref_kvs)

        self.kvs = kvs
        self.ref_kvs = ref_kvs

        ''' initialize SecCSLite '''
        self.seccs = seccs_class(
            self.S, kvs, crypto_wrapper, reference_counter=seccs.rc.DatabaseReferenceCounter(ref_kvs))
        self.R = self.seccs._R

    def tearDown(self):
        ''' delete references to data structure and force garbage collection '''
        del self.kvs_size
        del self.kvs_count
        del self.rc_count
        del self.kvs
        del self.ref_kvs
        del self.seccs
        del self.R
        del self.digest_size

        import gc
        gc.collect()

    def test_unsupported_chunksize(self):
        kvs = dict()
        crypto_wrapper = seccs.crypto_wrapper.SHA_256()
        self.assertRaises(
            seccs.UnsupportedChunkSizeError, seccs.SecCSLite, 63, kvs, crypto_wrapper)

    '''
    G1: The (expected) increase of the data structure's storage consumption caused by PutContent(m) should be in O(|m|).
    '''

    # @unittest.skip("not now")
    def test_small_inserts(self):
        ''' data structure should be empty at the beginning '''
        kvs_size = self.kvs_size()
        self.assertEqual(kvs_size, 0)
        kvs_count = self.kvs_count()
        self.assertEqual(kvs_count, 0)
        rc_count = self.rc_count()
        self.assertEqual(rc_count, 0)

        ''' verify that contents up to the average chunk size are stored in a single chunk '''
        for content_length in range(0, self.S + 1):
            content = bytes([self.random.randint(0, 255)
                             for _ in range(content_length)])
            self.seccs.put_content(content)

            kvs_size += content_length + self.digest_size
            kvs_count += 1
            rc_count += 1

            self.assertEqual(self.kvs_size(), kvs_size)
            self.assertEqual(self.kvs_count(), kvs_count)
            self.assertEqual(self.rc_count(), rc_count)

        ''' verify that contents beyond the average chunk size are stored in more than one chunk '''
        content = b''.join(bytes([self.random.randint(0, 255)])
                           for _ in range(self.S + 1))
        self.seccs.put_content(content)

        kvs_size += content_length + self.digest_size
        kvs_count += 1
        rc_count += 1

        self.assertGreater(self.kvs_size(), kvs_size)
        self.assertGreater(self.kvs_count(), kvs_count)
        self.assertGreater(self.rc_count(), rc_count)

    # @unittest.skip("not now")
    def test_large_inserts(self):
        ''' data structure should be empty at the beginning '''
        kvs_size = self.kvs_size()
        self.assertEqual(kvs_size, 0)
        kvs_count = self.kvs_count()
        self.assertEqual(kvs_count, 0)
        rc_count = self.rc_count()
        self.assertEqual(rc_count, 0)

        ''' verify for content lengths of different orders of magnitude '''
        for chunking_levels in range(1, 8):
            content_length = self.random.randint(
                int(self.S * (self.S / self.R) ** (chunking_levels - 1)), int(self.S * (self.S / self.R) ** (chunking_levels)))

            ''' averaged over 10 insertions, it is unlikely that a content's storage costs are above four times of its expected storage costs '''
            expected_costs = (content_length / self.S) * (self.S + 
                                                          self.digest_size) * 2  # storage for leaf chunks times two
            for _ in range(10):
                content = bytes([self.random.randint(0, 255)
                                 for _ in range(content_length)])
                self.seccs.put_content(content)

                kvs_size += 4 * expected_costs

            self.assertLess(self.kvs_size(), kvs_size)
            kvs_size = self.kvs_size()

    '''
    G2: If m is highly redundant, i.e. there exists another m' in the content store that
        is identical to m except for a single sequence of d bytes, the (expected) increase
        in storage consumption caused by PutContent(m) should be in O(d + log|m|).
    '''

    # @unittest.skip("not now")
    def test_similar_contents(self):
        ''' data structure should be empty at the beginning '''
        kvs_size = self.kvs_size()
        self.assertEqual(kvs_size, 0)
        kvs_count = self.kvs_count()
        self.assertEqual(kvs_count, 0)
        rc_count = self.rc_count()
        self.assertEqual(rc_count, 0)

        ''' verify for content lengths of different orders of magnitude '''
        for chunking_levels in range(1, 8):
            content_length = self.random.randint(
                int(self.S * (self.S / self.R) ** (chunking_levels - 1)), int(self.S * (self.S / self.R) ** (chunking_levels)))

            ''' choose random offset '''
            offset = self.random.randint(0, content_length)

            ''' choose random d '''
            d = self.random.randint(
                1, content_length // 10 + 1)  # should be at least an order of magnitude less than the length

            increase_size = 0
            expected_increase = 0

            ''' average over 10 runs '''
            for _ in range(10):
                content = bytes([self.random.randint(0, 255)
                                 for _ in range(content_length)])
                self.seccs.put_content(content)
                kvs_size = self.kvs_size()

                content = b''.join([content[:offset], bytes(
                    [self.random.randint(0, 255) for _ in range(d)]), content[offset + d:]])
                self.seccs.put_content(content)

                increase_size += self.kvs_size() - kvs_size

                ''' calculate expected increase '''
                expected_increase += 4 * \
                    ((math.ceil(math.log(content_length / self.S) / 
                                math.log(self.S / self.R)) + 1) * (self.S + self.digest_size) + d)

            self.assertLess(increase_size, expected_increase)
            kvs_size = self.kvs_size()

    '''
    G3: If any call k = PutContent(m) has been issued before m2 = GetContent(k), then it holds m2 = m or an Exception is raised.
    '''

    # @unittest.skip("not now")
    def test_small_retrievals(self):
        ''' verify contents up to the average chunk size that are stored in a single chunk '''
        for content_length in range(0, self.S + 1):
            content = bytes([self.random.randint(0, 255)
                             for _ in range(content_length)])
            k = self.seccs.put_content(content)

            m = self.seccs.get_content(k)
            self.assertEqual(m, content)

    # @unittest.skip("not now")
    def test_large_retrievals(self):
        ''' verify for content lengths of different orders of magnitude '''
        for chunking_levels in range(1, 8):
            content_length = self.random.randint(
                int(self.S * (self.S / self.R) ** (chunking_levels - 1)), int(self.S * (self.S / self.R) ** (chunking_levels)))

            content = bytes([self.random.randint(0, 255)
                             for _ in range(content_length)])
            k = self.seccs.put_content(content)

            m = self.seccs.get_content(k)
            self.assertEqual(m, content)

    '''
    G4: After DeleteContent(k) has been executed, the content store shall be in
    the same state as it would have been if neither that nor the most recent
    PutContent call that returned k had been made.
    '''

    # @unittest.skip("not now")
    def test_simple_immediate_deletion(self):
        ''' verify (insert a, delete a) vs. () '''

        ''' record data structure state '''
        kvs = dict(self.kvs)
        ref_kvs = dict(self.ref_kvs)

        ''' insert and delete contents of lengths of different orders of magnitude '''
        for chunking_levels in range(1, 8):
            content_length = self.random.randint(
                int(self.S * (self.S / self.R) ** (chunking_levels - 1)), int(self.S * (self.S / self.R) ** (chunking_levels)))

            content = bytes([self.random.randint(0, 255)
                             for _ in range(content_length)])

            ''' insert and delete immediately afterwards '''
            k = self.seccs.put_content(content)
            self.seccs.delete_content(k)

            self.assertEqual(self.kvs, kvs)
            self.assertEqual(self.ref_kvs, ref_kvs)

    # @unittest.skip("not now")
    def test_simple_nonimmediate_deletion(self):
        ''' verify (insert a) vs. (insert b, insert a, delete b) '''

        kvs_empty = dict(self.kvs)
        ref_kvs_empty = dict(self.ref_kvs)

        ''' insert some content and record data structure state '''
        content_a = bytes([self.random.randint(0, 255)
                           for _ in range(2 * 1024 * 1024)])
        k_a = self.seccs.put_content(content_a)

        kvs_with_a = dict(self.kvs)
        ref_kvs_with_a = dict(self.ref_kvs)

        ''' delete other content and ensure that data structure is empty again '''
        self.seccs.delete_content(k_a)
        self.assertEqual(self.kvs, kvs_empty)
        self.assertEqual(self.ref_kvs, ref_kvs_empty)

        ''' insert some content b, then a, then delete b '''
        content_b = bytes([self.random.randint(0, 255)
                           for _ in range(2 * 1024 * 1024)])
        k_b = self.seccs.put_content(content_b)

        self.seccs.put_content(content_a)

        self.seccs.delete_content(k_b)

        ''' data structure must be in the same state as it would have been if only a was inserted '''
        self.assertEqual(self.kvs, kvs_with_a)
        self.assertEqual(self.ref_kvs, ref_kvs_with_a)

    # @unittest.skip("not now")
    def test_complex_nonimmediate_deletion(self):
        ''' verify (insert a) vs. (insert b, insert a, delete b) with b being a slightly modified version of a '''

        kvs_empty = dict(self.kvs)
        ref_kvs_empty = dict(self.ref_kvs)

        ''' insert some content and record data structure state '''
        content_a = bytes([self.random.randint(0, 255)
                           for _ in range(2 * 1024 * 1024)])
        k_a = self.seccs.put_content(content_a)

        kvs_with_a = dict(self.kvs)
        ref_kvs_with_a = dict(self.ref_kvs)

        ''' delete other content and ensure that data structure is empty again '''
        self.seccs.delete_content(k_a)
        self.assertEqual(self.kvs, kvs_empty)
        self.assertEqual(self.ref_kvs, ref_kvs_empty)

        ''' insert some content b, then a, then delete b '''
        offset = self.random.randint(0, len(content_a))
        d = self.random.randint(1, len(content_a) // 10 + 1)
        content_b = b''.join([content_a[:offset], bytes(
            [self.random.randint(0, 255) for _ in range(d)]), content_a[offset + d:]])
        k_b = self.seccs.put_content(content_b)

        self.seccs.put_content(content_a)

        self.seccs.delete_content(k_b)

        ''' data structure must be in the same state as it would have been if only a was inserted '''
        self.assertEqual(self.kvs, kvs_with_a)
        self.assertEqual(self.ref_kvs, ref_kvs_with_a)

    # @unittest.skip("not now")
    def test_multiple_deletion(self):
        ''' verify (insert a, insert a, delete a, delete a) vs. () '''

        ''' record data structure state '''
        kvs = dict(self.kvs)
        ref_kvs = dict(self.ref_kvs)

        ''' insert and delete contents of lengths of different orders of magnitude '''
        for chunking_levels in range(1, 8):
            content_length = self.random.randint(
                int(self.S * (self.S / self.R) ** (chunking_levels - 1)), int(self.S * (self.S / self.R) ** (chunking_levels)))

            content = bytes([self.random.randint(0, 255)
                             for _ in range(content_length)])

            ''' insert and delete immediately afterwards '''
            k = self.seccs.put_content(content)
            k2 = self.seccs.put_content(content)

            self.assertEqual(k, k2)

            self.seccs.delete_content(k)
            self.seccs.delete_content(k2)

            self.assertEqual(self.kvs, kvs)
            self.assertEqual(self.ref_kvs, ref_kvs)

    # @unittest.skip("not now")
    def test_deletion_after_multiple_insertion(self):
        ''' verify (insert a, insert a, delete a) vs. (insert_a) '''

        kvs_empty = dict(self.kvs)
        ref_kvs_empty = dict(self.ref_kvs)

        ''' insert some content and record data structure state '''
        content_a = bytes([self.random.randint(0, 255)
                           for _ in range(2 * 1024 * 1024)])
        k_a = self.seccs.put_content(content_a)

        kvs_with_a = dict(self.kvs)
        ref_kvs_with_a = dict(self.ref_kvs)

        ''' delete other content and ensure that data structure is empty again '''
        self.seccs.delete_content(k_a)
        self.assertEqual(self.kvs, kvs_empty)
        self.assertEqual(self.ref_kvs, ref_kvs_empty)

        ''' insert some content a, then a again, then delete a '''
        k_a = self.seccs.put_content(content_a)

        k_a2 = self.seccs.put_content(content_a)
        self.seccs.delete_content(k_a2)

        ''' data structure must be in the same state as it would have been if only a was inserted '''
        self.assertEqual(self.kvs, kvs_with_a)
        self.assertEqual(self.ref_kvs, ref_kvs_with_a)

if __name__ == "__main__":

    ''' list of chunk sizes that should be evaluated '''
    CHUNK_SIZES = [128]

    ''' list of seeds that triggered an error in the past '''
    REQUIRED_SEEDS = [
        float.fromhex('0x1.81c6d350b9a20p-6'),
        float.fromhex('0x1.6221d735fc060p-1'),
        float.fromhex('0x1.1e20ca763865ep-1'),
        float.fromhex('0x1.c17016a32c630p-3'),
        float.fromhex('0x1.c9dd7b04184f8p-2'),
        float.fromhex('0x1.1431bafbf716ap-1'),
        float.fromhex('0x1.7b5da4558c9dep-2')
    ]

    ''' number of seeds to be chosen randomly '''
    RANDOM_TESTS_COUNT = 1

    ''' execute all tests with required seeds and RANDOM_TESTS_COUNT randomly chosen seeds '''
    suite = unittest.TestSuite()
    for cls in [RCTest, CryptoWrapperTest]:
        for name in unittest.TestLoader().getTestCaseNames(cls):
            suite.addTest(cls(name))
    
#     # perform one test with leaf padding crypto wrapper
#     for S in CHUNK_SIZES:
#         for name in unittest.TestLoader().getTestCaseNames(SecCSLiteTest):
#             suite.addTest(SecCSLiteTest(name, S=S, crypto_wrapper=seccs.crypto_wrapper.HMAC_SHA_256_DISTINGUISHED_ROOT_WITH_LEAF_PADDING(b'0000000000000000')))
    
    for seed in REQUIRED_SEEDS + [None] * RANDOM_TESTS_COUNT:
        for S in CHUNK_SIZES:
            for name in unittest.TestLoader().getTestCaseNames(SecCSLiteTest):
                suite.addTest(SecCSLiteTest(name, S=S, seed=seed)
                              if seed is not None else SecCSLiteTest(name, S=S))
    unittest.TextTestRunner(verbosity=2, failfast=True).run(suite)



"""sec-cs --- the SECure Content Store.

This module provides an implementation of the secure content store data
structure introduced in [LS16]_.

`sec-cs` allows secure and efficient storage of contents in an existing
key-value database, providing the following features:

    * Confidentiality:
        Stored contents are securely encrypted using a symmetric key.
    * Authenticity:
        `sec-cs` guarantees authenticity of all stored contents,
        irrespective of gurantees of the underlying database.
    * Storage Efficiency:
        Data deduplication strategies are applied to all
        stored contents. When storing new contents, overlapping parts of
        existing contents are automatically reused as to avoid redundancy.
        Storage costs of an n-bytes content that differs only slightly from an
        existing content are in O(log n).
        
Note:
    The only `sec-cs` implementation currently included in this module is called
    :class:`SecCSLite`. While it is likely suitable for many projects and can be
    used as is, it is actually intended as a base class for a much more powerful
    variant, :class:`SecCS`, which makes some slight changes to the internal
    storage structure and will be published in the near future.

References:
    .. [LS16] Dominik Leibenger and Christoph Sorge (2016). sec-cs: Getting the
       Most out of Untrusted Cloud Storage.
       `arXiv:1606.03368 <http://arxiv.org/abs/1606.03368>`_
"""
import collections
from itertools import chain
import itertools
import logging
import math
import struct

import seccs.crypto_wrapper
import seccs.rc

__version__ = '0.0.5'


class UnsupportedChunkSizeError(Exception):
    """Raised when trying to instantiate SecCS with an unsupported chunk size."""
    pass


class SecCSLite(object):

    """Secure Content Store `lite`.

    Basic implementation of the Secure Content Store data structure, supports
    only insertion (put), retrieval (get) and deletion (delete) of contents.

    Args:
        chunk_size (int): Target chunk size, i.e., expected size of all chunks
            stored in the database.
        database: Persistent database used as backend. Can be any object with
            a dict-like interface, i.e., any object implementing the operations
            __getitem__, __setitem__, __delitem__ and __contains__.
        crypto_wrapper (:class:`crypto_wrapper.BaseCryptoWrapper`): Crypto
            wrapper object that specifies cryptographic operations to be applied
            to data stored in the `database`.
        chunking_strategy (Optional[:class:`fastchunking.BaseChunkingStrategy`]):
            Chunking strategy that shall be applied to contents. Defaults to
            Rabin-Karp-based content defined chunking with 48-bytes window size.
        reference_counter (Optional[:class:`seccs.rc.BaseReferenceCounter`]):
            Reference counting strategy. By default, reference counters are
            stored in `database` under keys `key` || `"r"`, where `key` is the key
            whose references are counted.
        **kwargs: Extra keyword arguments that you should NOT use unless you
            really, really know what you are doing, e.g.:

            * length_to_height_fn: Function that resolves content lengths to
              appropriate chunk tree heights. May be used to modify the
              multi-level chunking approach performed by default, e.g., to
              degrade it to single-level chunking or similar.
            * height_to_chunk_size_fn: Function that computes the target chunk
              size for a specific level (height) of a chunk tree. May be used
              to create `imbalanced` chunk trees.

    Raises:
        seccs.UnsupportedChunkSizeError: If the chosen chunk size would create
            superchunk nodes with less than two expected children as efficiency
            guarantees would fail in this case (see [LS16]_).
    """

    def __init__(
            self, chunk_size, database, crypto_wrapper, chunking_strategy=None,
            reference_counter=None, **kwargs):
        self._logger = logging.getLogger(__name__)

        """ initialize reference representations depending on digest size """
        self._digest_size = crypto_wrapper.DIGEST_SIZE
        self._initialize_reference_representations()

        """ require S >= 2R unless an alternate height_to_chunk_size_fn is given """
        if chunk_size < 2 * self._R and 'height_to_chunk_size_fn' not in kwargs:
            raise UnsupportedChunkSizeError((
                'target average chunk size of {S} bytes is too small, must be '
                'at least {req} bytes').format(S=chunk_size, req=2 * self._R))

        """ function that calculates the appropriate number of chunking levels
            for a specific content length """
        self._content_length_to_level = kwargs.get('length_to_height_fn', None)
        if not self._content_length_to_level:
            def length_to_height_fn(size):
                return int(max(math.ceil(math.log(size * 1.0 / chunk_size)
                                         / math.log(chunk_size * 1.0 / self._R)),
                               0)) if size != 0 else 0
            self._content_length_to_level = length_to_height_fn

        """ dict that resolves a chunking level to its corresponding chunk size """
        self._level_to_chunksize = self._LevelToChunkSizeDict(
            chunk_size, self._R, kwargs.get('height_to_chunk_size_fn', None))

        """
        Crypto wrapper.

          Expected interface:
            value, digest <- wrap_value(value, height, is_root)
            value <- unwrap_value(value, digest, height, is_root, length)
        """
        self._crypto_wrapper = crypto_wrapper

        """
        Key-value database.

          Expected interface:
            __setitem__(key, value)
            __delitem__(key)
            value <- __getitem__(key)
            True/False <- __contains__(key)
        """
        self._database = database

        """
        Reference counter.

          Expected interface:
            new_count <- inc(key)
            new_count <- dec(key)
        """
        if reference_counter is None:
            reference_counter = seccs.rc.KeySuffixDatabaseReferenceCounter(
                database, b'r')
        self._reference_counter = reference_counter

        """
        Chunking strategy.

        Needs to be a context-sensitive chunking strategy meeting the Chunking
        interface of the fastchunking package.
        """
        if chunking_strategy is None:
            import fastchunking
            chunking_strategy = fastchunking.RabinKarpCDC(
                48, seed=0)  # FIXME: seed should not be 0
        self._chunking_strategy = chunking_strategy
        self._W = chunking_strategy.window_size

    def _initialize_reference_representations(self):
        """Define how chunk references are represented."""

        # size of subchunk references
        digest_size = self._digest_size
        self._R = digest_size

        # format of content references
        self._content_reference_format = '!{digest_size}sQ'.format(
            digest_size=digest_size)

        # format of subchunk references
        self._subchunk_reference_format = '!{digest_size}s'.format(
            digest_size=digest_size)

    """
    Public interface.
    """

    def put_content(self, m, ignore_rc=False):
        """Insert a content into the data structure.

        Args:
            m (str): The message or content that shall be processed and inserted
                into the data structure.
            ignore_rc (Optional[bool]): If True, increase of reference counter
                for the root node of the generated chunk tree is skipped.
                Defaults to False.

        Returns:
            str: Digest of the content that allows its retrieval using
            :meth:`get_content`.
        """
        return self.put_content_and_check_if_new(m, ignore_rc)[0]

    def put_content_and_check_if_new(self, m, ignore_rc=False):
        """Insert a content into the data structure.

        Like :meth:`put_content`, but return value includes information whether
        the content had been in the data structure before.

        Args:
            m (str): The message or content that shall be processed and inserted
                into the data structure.
            ignore_rc (Optional[bool]): If True, increase of reference counter
                for the root node of the generated chunk tree is skipped.
                Defaults to False.

        Returns:
            tuple: (digest, is_new), where `digest` is the content's digest that
            allows its retrieval using :meth:`get_content`, and `is_new`
            is True if the content has been inserted for the first time and
            False if it had existed before.
        """
        l = len(m)
        h = self._content_length_to_level(l)
        k, is_new = self._put_chunk(m, h, h)
        k = k[0]
        if not ignore_rc:
            self._reference_counter.inc(k)
        return (struct.pack(self._content_reference_format, k, l), is_new)

    def get_content(self, k):
        """Retrieve a content from the data structure.

        Args:
            k (str): The digest under which the content is stored.

        Returns:
            str: The content bytestring.
        """
        k, l = struct.unpack(self._content_reference_format, k)
        h = self._content_length_to_level(l)
        return self._get_chunk(k, h, h, l)

    def delete_content(self, k, ignore_rc=False):
        """Delete a content from the data structure.

        Decreases the content's reference counter and deletes its root chunk
        (possibly including children) if no references are left.

        Args:
            k (str): The digest under which the content is stored.
            ignore_rc (Optional[bool]): If True, decrease of reference counter
                of the root node is skipped and root node (possibly including
                children) is deleted straight away.
        """
        k, l = struct.unpack(self._content_reference_format, k)
        h = self._content_length_to_level(l)
        return self._delete_content(k, h, h, l, ignore_rc)

    """
    Helper functions for storing chunk tree nodes in the backend and retrieving
    them without having to deal with (de)serialization issues.
    """

    def _store_node(self, v, h, root_h):
        """Store chunk tree node in database.

        Args:
            v: Node content (str in case of a leaf chunk, list of child
                references otherwise).
            h (int): Height of chunk tree node.
            root_h (int): Height of chunk tree.

        Returns:
            tuple: ((k, ), is_new), where `k` is the digest or key under which
                the node has been stored, and `is_new` is a flag indicating
                whether the node has just been inserted (True) or existed before
                (False).
        """

        if isinstance(v, list):
            """ serialize superchunk:
                use each key as is (and replace None keys by a sequence of zero
                bytes) and concatenate everything """
            serialized_chunk = b''.join([struct.pack(
                self._subchunk_reference_format,
                subchunk_key) for (subchunk_key,) in v])

        else:
            """ leaf chunks do not need any serialization """
            serialized_chunk = v

        """ compute chunk key using digest function """
        serialized_chunk, k = self._crypto_wrapper.wrap_value(
            serialized_chunk, h, h == root_h)

        """ if the chunk already exists, verify its integrity and return its identifier """
        if k in self._database and self._get_node(k, h, root_h) is not None and self._reference_counter.get(k) > 0:
            return ((k,), False)

        """ if we have a new superchunk, increase reference counters for its children """
        if isinstance(v, list):
            for (chunk_id,) in v:
                self._reference_counter.inc(chunk_id)

        """ create the chunk and return its identifier """
        self._database[k] = serialized_chunk

        return ((k,), True)

    def _get_node(self, k, h, root_h, l=-1):
        """Retrieve chunk tree node from database.

        Args:
            k (str): Digest or key of the node.
            h (int): Height of the chunk tree node in its chunk tree.
            root_h (int): Height of the full chunk tree.
            l [Optional(int)]: Size of the node representation, if known.
                Defaults to -1.

        Returns:
            Chunk tree node, i.e., a bytestring in case of a leaf chunk (h = 0),
            and a list of subchunk references in case of a superchunk (h > 0).
        """

        v = self._database[k]

        serialized_chunk = self._crypto_wrapper.unwrap_value(
            v, k, h, h == root_h, l)

        if h > 0:
            """ deserialize superchunk """
            return None if serialized_chunk is None else[
                struct.unpack(
                    self._subchunk_reference_format,
                    serialized_chunk[i: i + self._R])
                for i in range(0, len(serialized_chunk),
                               self._R)]
        else:
            return serialized_chunk

    """
    Functions for realization of operations.
    """

    def _put_chunk(self, m, h, root_h):
        """ Compute subtree of a chunk tree for a specific content part and
        store it in the database.

        Args:
            m (str): Content part (message).
            h (int): Height of the considered content part in its content's
                chunk tree.
            root_h (int): Height of chunk tree of the whole content.

        Returns:
            tuple: ((k, ), is_new), where `k` is the digest or key under which
                the chunk has been stored, and `is_new` is a flag indicating
                whether the chunk has just been inserted (True) or existed
                before (False).
        """
        assert h >= 0
        if __debug__:
            self._logger.debug(
                'Store chunk with length %d bytes (levels: %d)', len(m), h)

        if h == 0:
            """ No chunking required """
            return self._store_node(m, h, root_h)  # return (k), is_new

        else:
            """ Chunking required """
            if __debug__:
                self._logger.debug(
                    'Do chunking with target chunk size: %d', self._level_to_chunksize[h - 1])

            """ Determine chunk boundaries for all levels """

            # first boundary is before the content
            boundaries = [(0, h - 1)]

            # determine subsequent boundaries...
            chunker = self._chunking_strategy.create_multilevel_chunker(
                [self._level_to_chunksize[height] for height in range(0, h)])
            # ...and append them to the boundaries list
            # W-1 zeros are prepended
            boundaries.extend(
                chunker.next_chunk_boundaries_levels(m, self._W - 1))

            # make sure that the last boundary is after the content and is a
            # height (h-1)-boundary
            if boundaries[-1][0] == len(m):
                boundaries.pop()
            boundaries.append((len(m), h - 1))

            """ Build chunk tree and store its nodes """

            # initialize temporary nodes (one for each level)
            nodes_levels = dict([(height, []) for height in range(0, h + 1)])

            # process chunks specified by boundaries
            for ((start_position, _), (end_position, boundary_height)) \
                    in zip(boundaries, boundaries[1:]):
                # place leaf chunk at height 0
                # is guaranteed to be cleared during this loop due to the
                # subsequent inner loop
                nodes_levels[0] = m[start_position:end_position]

                # update superchunks and insert nodes that are already complete
                # (including the leaf chunk)
                for height in range(0, boundary_height + 1):
                    # in order to extend the superchunk, we have to insert its
                    # next subchunk
                    nodes_levels[height + 1].append(self._store_node(nodes_levels[height],
                                                                     height, root_h)[0])
                    # subchunk is complete, so prepare new subchunk
                    nodes_levels[height] = []

            # insert root chunk (return (k), is_new)
            return self._store_node(nodes_levels[h], h, root_h)

    def _get_chunk(self, k, h, root_h, l):
        """ Get content part which is stored under some key in the database.
        If content consists of multiple chunks, put them together.

        Args:
            k (str): Digest or key of the node representing the part of the
                chunk tree that is to be retrieved.
            h (int): Height of the node representing the part of the chunk tree
                that is to be retrieved.
            root_h (int): Height of the chunk tree to which the chunk requested
                here belongs.
            l (int): Length of the content represented by the chunk that is to
                be retrieved.

        Returns:
            Content part.
        """
        chunks = [(k,)]
        for height in range(h, 0, -1):
            chunks = chain.from_iterable([self._get_node(k, height, root_h)
                                          for (k,) in chunks])
        return b''.join([self._get_node(k, 0, root_h) for (k,) in chunks])

    def _delete_content(self, k, h, root_h, l, ignore_rc=True):
        """ Remove reference from content's root node and delete its root node
        (possibly including children) if necessary.

        Args:
            k (str): The digest under which the content is stored.
            h (int): Height of node `k` in its chunk tree.
            root_h (int): Height of the chunk tree `k` is part of.
            l (int): Length of the content represented by `k`.
            ignore_rc (Optional[bool]): If True, decrease of reference counter
                of the root node is skipped and root node (possibly including
                children) is deleted straight away.
        """
        if ignore_rc or self._reference_counter.dec(k) == 0:
            return self._delete_chunk(k, h, root_h, l)

    def _delete_chunk(self, k, h, root_h, l=-1):
        """ Delete a chunk and all children whose reference counters drop to
        zero.

        Args:
            k (str): The digest of the chunk tree node that is to be deleted.
            h (int): The height of the chunk tree node `k`.
            root_h (int): The height of the chunk tree `k` is part of.
            l [Optional(int)]: The length of the content part represented by
                `k`. Defaults to -1.
        """
        if h > 0:
            v = self._get_node(k, h, root_h)
            for (k_child,) in v:
                if self._reference_counter.dec(k_child) == 0:
                    self._delete_chunk(k_child, h - 1, root_h)
        del self._database[k]

    """
    Generic helpers.
    """

    class _LevelToChunkSizeDict(collections.defaultdict):

        """
        Dict that caches the target chunk sizes for different chunk tree heights
        to save some computations.
        """

        def __init__(self, target_chunksize, R, height_to_chunk_size_fn=None):
            collections.defaultdict.__init__(self)
            self._target_chunksize = target_chunksize
            self._R = R

            if not height_to_chunk_size_fn:
                height_to_chunk_size_fn = lambda L: int(
                    self._target_chunksize ** (L + 1) / self._R ** L)
            self._height_to_chunk_size_fn = height_to_chunk_size_fn

        def __missing__(self, L):
            self[L] = chunksize = self._height_to_chunk_size_fn(L)
            return chunksize



"""Simple reference counter implementations."""
import abc


class BaseReferenceCounter(object):

    """Abstract base class for reference counters."""
    __metaclass__ = abc.ABCMeta
    __slots__ = []

    def inc(self, key):
        """Abstract increment interface.

        Args:
            key: Key whose reference counter shall be incremented.

        Returns:
            Number of references of key after increment.
        """
        raise NotImplementedError

    def dec(self, key):
        """Abstract decrement interface.

        Args:
            key: Key whose reference counter shall be decremented.

        Returns:
            Number of references of key after decrement.
        """
        raise NotImplementedError

    def get(self, key):
        """Abstract get interface.

        Args:
            key: Key whose reference counter shall be retrieved.

        Returns:
            Number of references of key.
        """
        raise NotImplementedError


class NoReferenceCounter(BaseReferenceCounter):

    """Non-counting reference counter, always returns 1 for any key.

    Can be used to disable reference counting where a reference counter is
    required.
    """

    __slots__ = []

    def inc(self, key):
        """Increment interface.

        Returns:
            1
        """
        return 1

    def dec(self, key):
        """Decrement interface.

        Returns:
            1
        """
        return 1

    def get(self, key):
        """Get interface.

        Returns:
            1
        """
        return 1


class DatabaseReferenceCounter(BaseReferenceCounter):

    """Database-backed reference counter.

    Uses a given database to store reference counters. The reference counter of
    an element `key` is stored as follows:
    
        * If its value is 0, `key` is not stored in the database.
        * If its value is > 0, its int value is stored in the database under
          `key`.

    Args:
        database: Database object with a dict-like interface, i.e., implementing
            the operations __getitem__, __setitem__ and __delitem__.
    """

    __slots__ = ['_database']

    def __init__(self, database):
        super(DatabaseReferenceCounter, self).__init__()

        self._database = database

    def inc(self, key):
        """Incrementes reference counter of key.

        See :meth:`.BaseReferenceCounter.inc`.
        """
        database = self._database
        database[key] = new_count = (
            database[key] if key in database else 0) + 1
        return new_count

    def dec(self, key):
        """Decrementes reference counter of key.

        See :meth:`.BaseReferenceCounter.dec`.
        """
        database = self._database
        new_count = database[key] - 1
        if new_count == 0:
            del database[key]
        else:
            database[key] = new_count
        return new_count

    def get(self, key):
        """Gets reference counter of key.

        See :meth:`.BaseReferenceCounter.get`.
        """
        database = self._database
        if key in database:
            return database[key]
        else:
            return 0


class KeySuffixDatabaseReferenceCounter(DatabaseReferenceCounter):

    """Database-backed reference counter.

    Similar to :class:`.DatabaseReferenceCounter`, but the reference counter of
    a `key` is not stored directly under `key`, but under `key` || `suffix`.

    Args:
        database: Database object with a dict-like interface, i.e., implementing
            the operations __getitem__, __setitem__ and __delitem__.
        suffix: Suffix for keys.
    """

    __slots__ = ['_suffix']

    def __init__(self, database, suffix):
        super(KeySuffixDatabaseReferenceCounter, self).__init__(database)

        self._suffix = suffix

    def inc(self, key):
        """Incrementes reference counter of key.

        See :meth:`.DatabaseReferenceCounter.inc`.
        """
        return DatabaseReferenceCounter.inc(self, key + self._suffix)

    def dec(self, key):
        """Decrementes reference counter of key.

        See :meth:`.DatabaseReferenceCounter.dec`.
        """
        return DatabaseReferenceCounter.dec(self, key + self._suffix)

    def get(self, key):
        """Gets reference counter of key.

        See :meth:`.DatabaseReferenceCounter.get`.
        """
        return DatabaseReferenceCounter.get(self, key + self._suffix)



import codecs
import os
import re
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

# Some general-purpose code stolen from
# https://github.com/jeffknupp/sandman/blob/5c4b7074e8ba5a60b00659760e222c57ad24ef91/setup.py

here = os.path.abspath(os.path.dirname(__file__))


class Tox(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)


def read(*parts):
    # intentionally *not* adding an encoding option to open
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='seccs',
    version=find_version('seccs', '__init__.py'),

    description='Secure Content Store',
    long_description=open('README.rst').read(),

    url='https://github.com/netleibi/seccs',

    author='Dominik Leibenger',
    author_email='python-seccs@mails.dominik-leibenger.de',

    license='Apache Software License',

    classifiers=[
            'Development Status :: 2 - Pre-Alpha',

            'Intended Audience :: Developers',
            'Topic :: Software Development :: Libraries :: Python Modules',

            'License :: OSI Approved :: Apache Software License',

            'Operating System :: OS Independent',

            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5'
    ],

    packages=['seccs'],

    install_requires=['fastchunking'],

    test_suite='seccs.test',
    tests_require=['tox'],
    cmdclass={'test': Tox}
)
